@@ -7,7 +7,7 @@ use vars qw($VERSION @ISA @EXPORT @EXPORT_OK $AUTOLOAD);
 
 require Exporter;
 
-$VERSION = '0.15';
+$VERSION = '0.16';
 @ISA = qw(Exporter);
 @EXPORT_OK = qw (
   can_log
@@ -57,6 +57,7 @@ use File::Path; # mkpath
 use File::Copy;
 use Carp;
 use Storable;
+use IPC::Cmd;
 
 my $have_mime_lite;
 eval {
@@ -357,594 +358,10 @@ sub disable_all_signals {
   }
 }
 
-# incompatible with POSIX::SigAction
-#
-sub install_layered_signal {
-  my ($s, $handler_code) = @_;
-
-  my %available_signals = map {$_ => 1} keys %SIG;
-
-  Yandex::Tools::die("install_layered_signal got nonexistent signal name [$s]")
-    unless defined($available_signals{$s});
-  Yandex::Tools::die("install_layered_signal expects coderef")
-    if !ref($handler_code) || ref($handler_code) ne 'CODE';
-
-  my $previous_handler = $SIG{$s};
-
-  my $sig_handler = sub {
-    my ($called_sig_name, @sig_param) = @_;
-    
-    # $s is a closure referring to real signal name
-    # for which this handler is being installed.
-    # it is used to distinguish between
-    # real signal handlers and aliased signal handlers
-    my $signal_name = $s;
-
-    # $called_sig_name is a signal name which
-    # was passed to this signal handler;
-    # it doesn't equal $signal_name in case
-    # some signal handlers in %SIG point
-    # to other signal handler (CHLD and CLD,
-    # ABRT and IOT)
-    #
-    # initial signal handler for aliased signal
-    # calls some other signal handler which
-    # should not execute the same handler_code again
-    if ($called_sig_name eq $signal_name) {
-      $handler_code->($signal_name);
-    }
-
-    # run original signal handler if any (including aliased)
-    #
-    if (ref($previous_handler)) {
-      $previous_handler->($called_sig_name, @sig_param);
-    }
-  };
-
-  $SIG{$s} = $sig_handler;
-}
-
-# give process a chance sending TERM,
-# waiting for a while (2 seconds)
-# and killing it with KILL
-sub kill_gently {
-  my ($pid, $opts) = @_;
-  
-  $opts = {} unless $opts;
-  $opts->{'wait_time'} = 2 unless defined($opts->{'wait_time'});
-  $opts->{'first_kill_type'} = 'just_process' unless $opts->{'first_kill_type'};
-  $opts->{'final_kill_type'} = 'just_process' unless $opts->{'final_kill_type'};
-
-  if ($opts->{'first_kill_type'} eq 'just_process') {
-    kill(15, $pid);
-  }
-  elsif ($opts->{'first_kill_type'} eq 'process_group') {
-    kill(-15, $pid);
-  }
-  
-  my $child_finished = 0;
-  my $wait_start_time = clock_gettime(CLOCK_MONOTONIC);
-
-  while (!$child_finished && $wait_start_time + $opts->{'wait_time'} > clock_gettime(CLOCK_MONOTONIC)) {
-    my $waitpid = waitpid($pid, WNOHANG);
-    if ($waitpid eq -1) {
-      $child_finished = 1;
-    }
-    Time::HiRes::usleep(250000); # quarter of a second
-  }
-
-  if (!$child_finished) {
-    if ($opts->{'final_kill_type'} eq 'just_process') {
-      kill(9, $pid);
-    }
-    elsif ($opts->{'final_kill_type'} eq 'process_group') {
-      kill(-9, $pid);
-    }
-  }
-}
-
-sub open3_run {
-  my ($cmd, $opts) = @_;
-
-  $opts = {} unless $opts;
-  
-  my $child_in = FileHandle->new;
-  my $child_out = FileHandle->new;
-  my $child_err = FileHandle->new;
-  $child_out->autoflush(1);
-  $child_err->autoflush(1);
-
-  my $pid = open3($child_in, $child_out, $child_err, $cmd);
-
-  # push my child's pid to our parent
-  # so in case i am killed parent
-  # could stop my child (search for
-  # child_child_pid in parent code)
-  if ($opts->{'parent_info'}) {
-    my $ps = $opts->{'parent_info'};
-    print $ps "spawned $pid\n";
-  }
-
-  if ($child_in && $child_out->opened && $opts->{'child_stdin'}) {
-
-    # If the child process dies for any reason,
-    # the next write to CHLD_IN is likely to generate
-    # a SIGPIPE in the parent, which is fatal by default.
-    # So you may wish to handle this signal.
-    #
-    # from http://perldoc.perl.org/IPC/Open3.html,
-    # absolutely needed to catch piped commands errors.
-    #
-    local $SIG{'PIPE'} = sub { 1; };
-    
-    print $child_in $opts->{'child_stdin'};
-  }
-  close($child_in);
-
-  my $child_output = {
-    'out' => $child_out->fileno,
-    'err' => $child_err->fileno,
-    $child_out->fileno => {
-      'parent_socket' => $opts->{'parent_stdout'},
-      'scalar_buffer' => "",
-      'child_handle' => $child_out,
-      'block_size' => ($child_out->stat)[11] || 1024,
-      },
-    $child_err->fileno => {
-      'parent_socket' => $opts->{'parent_stderr'},
-      'scalar_buffer' => "",
-      'child_handle' => $child_err,
-      'block_size' => ($child_err->stat)[11] || 1024,
-      },
-    };
-
-  my $select = IO::Select->new();
-  $select->add($child_out, $child_err);
-
-  # pass any signal to the child
-  # effectively creating process
-  # strongly attached to the child:
-  # it will terminate only after child
-  # has terminated (except for SIGKILL,
-  # which is specially handled)
-  foreach my $s (keys %SIG) {
-    my $sig_handler;
-    $sig_handler = sub {
-      kill("$s", $pid);
-      $SIG{$s} = $sig_handler;
-    };
-    $SIG{$s} = $sig_handler;
-  }
-
-  my $child_finished = 0;
-
-  my $got_sig_child = 0;
-  $SIG{'CHLD'} = sub { $got_sig_child = clock_gettime(CLOCK_MONOTONIC); };
-
-  while(!$child_finished && ($child_out->opened || $child_err->opened)) {
-
-    # parent was killed otherwise we would have got
-    # the same signal as parent and process it same way
-    if (getppid() eq "1") {
-      
-      # end my process group with all the children
-      # (i am the process group leader, so my pid
-      # equals to the process group id)
-      #
-      # same thing which is done
-      # with $opts->{'clean_up_children'}
-      # in run_forked
-      #
-      kill(-9, $$);
-
-      exit 1;
-    }
-
-    if ($got_sig_child) {
-      if (clock_gettime(CLOCK_MONOTONIC) - $got_sig_child > 1) {
-        # select->can_read doesn't return 0 after SIG_CHLD
-        #
-        # "On POSIX-compliant platforms, SIGCHLD is the signal
-        # sent to a process when a child process terminates."
-        # http://en.wikipedia.org/wiki/SIGCHLD
-        #
-        # nevertheless kill KILL wouldn't break anything here
-        #
-        kill (9, $pid);
-        $child_finished = 1;
-      }
-    }
-
-    Time::HiRes::usleep(1);
-
-    foreach my $fd ($select->can_read(1/100)) {
-      my $str = $child_output->{$fd->fileno};
-      Yandex::Tools::die("child stream not found: $fd") unless $str;
-
-      my $data;
-      my $count = $fd->sysread($data, $str->{'block_size'});
-
-      if ($count) {
-        if ($str->{'parent_socket'}) {
-          my $ph = $str->{'parent_socket'};
-          print $ph $data;
-        }
-        else {
-          $str->{'scalar_buffer'} .= $data;
-        }
-      }
-      elsif ($count eq 0) {
-        $select->remove($fd);
-        $fd->close();
-      }
-      else {
-        Yandex::Tools::die("error during sysread: " . $!);
-      }
-    }
-  }
-
-  my $waitpid_ret = waitpid($pid, 0);
-  my $real_exit = $?;
-  my $exit_value  = $real_exit >> 8;
-
-  # since we've successfully reaped the child,
-  # let our parent know about this.
-  #
-  if ($opts->{'parent_info'}) {
-    my $ps = $opts->{'parent_info'};
-
-    # child was killed, inform parent
-    if ($real_exit & 127) {
-      print $ps "$pid killed with " . ($real_exit & 127) . "\n";
-      
-    }
-
-    print $ps "reaped $pid\n";
-  }
-
-  if ($opts->{'parent_stdout'} || $opts->{'parent_stderr'}) {
-    return $exit_value;
-  }
-  else {
-    return {
-      'stdout' => $child_output->{$child_output->{'out'}}->{'scalar_buffer'},
-      'stderr' => $child_output->{$child_output->{'err'}}->{'scalar_buffer'},
-      'exit_code' => $exit_value,
-      };
-  }
-}
-
-sub run_forked {
-  my ($cmd, $opts) = @_;
-
-  $opts = {} unless $opts;
-  $opts->{'timeout'} = 0 unless $opts->{'timeout'};
-  $opts->{'terminate_wait_time'} = 2 unless defined($opts->{'terminate_wait_time'});
-
-  # turned on by default
-  $opts->{'clean_up_children'} = 1 unless defined($opts->{'clean_up_children'});
-
-  # sockets to pass child stdout to parent
-  my $child_stdout_socket;
-  my $parent_stdout_socket;
-
-  # sockets to pass child stderr to parent
-  my $child_stderr_socket;
-  my $parent_stderr_socket;
-  
-  # sockets for child -> parent internal communication
-  my $child_info_socket;
-  my $parent_info_socket;
-
-  socketpair($child_stdout_socket, $parent_stdout_socket, AF_UNIX, SOCK_STREAM, PF_UNSPEC) ||
-    Yandex::Tools::die ("socketpair: $!");
-  socketpair($child_stderr_socket, $parent_stderr_socket, AF_UNIX, SOCK_STREAM, PF_UNSPEC) ||
-    Yandex::Tools::die ("socketpair: $!");
-  socketpair($child_info_socket, $parent_info_socket, AF_UNIX, SOCK_STREAM, PF_UNSPEC) ||
-    Yandex::Tools::die ("socketpair: $!");
-
-  $child_stdout_socket->autoflush(1);
-  $parent_stdout_socket->autoflush(1);
-  $child_stderr_socket->autoflush(1);
-  $parent_stderr_socket->autoflush(1);
-  $child_info_socket->autoflush(1);
-  $parent_info_socket->autoflush(1);
-
-  my $start_time = clock_gettime(CLOCK_MONOTONIC);
-
-  my $pid;
-  if ($pid = fork) {
-
-    # we are a parent
-    close($parent_stdout_socket);
-    close($parent_stderr_socket);
-    close($parent_info_socket);
-
-    my $flags;
-
-    # prepare sockets to read from child
-
-    $flags = 0;
-    fcntl($child_stdout_socket, F_GETFL, $flags) || Yandex::Tools::die("can't fnctl F_GETFL: $!");
-    $flags |= O_NONBLOCK;
-    fcntl($child_stdout_socket, F_SETFL, $flags) || Yandex::Tools::die("can't fnctl F_SETFL: $!");
-
-    $flags = 0;
-    fcntl($child_stderr_socket, F_GETFL, $flags) || Yandex::Tools::die("can't fnctl F_GETFL: $!");
-    $flags |= O_NONBLOCK;
-    fcntl($child_stderr_socket, F_SETFL, $flags) || Yandex::Tools::die("can't fnctl F_SETFL: $!");
-
-    $flags = 0;
-    fcntl($child_info_socket, F_GETFL, $flags) || Yandex::Tools::die("can't fnctl F_GETFL: $!");
-    $flags |= O_NONBLOCK;
-    fcntl($child_info_socket, F_SETFL, $flags) || Yandex::Tools::die("can't fnctl F_SETFL: $!");
-
-#    print "child $pid started\n";
-
-    my $child_timedout = 0;
-    my $child_finished = 0;
-    my $child_stdout = '';
-    my $child_stderr = '';
-    my $child_merged = '';
-    my $child_exit_code = 0;
-    my $child_killed_by_signal = 0;
-    my $parent_died = 0;
-
-    my $got_sig_child = 0;
-    my $got_sig_quit = 0;
-    my $orig_sig_child = $SIG{'CHLD'};
-
-    $SIG{'CHLD'} = sub { $got_sig_child = clock_gettime(CLOCK_MONOTONIC); };
-    
-    if ($opts->{'terminate_on_signal'}) {
-      install_layered_signal($opts->{'terminate_on_signal'}, sub { $got_sig_quit = clock_gettime(CLOCK_MONOTONIC); });
-    }
-
-    my $child_child_pid;
-
-    while (!$child_finished) {
-      my $now = clock_gettime(CLOCK_MONOTONIC);
-
-      if ($opts->{'terminate_on_parent_sudden_death'}) {
-        $opts->{'runtime'}->{'last_parent_check'} = 0
-          unless defined($opts->{'runtime'}->{'last_parent_check'});
-
-        # check for parent once each five seconds
-        if ($now - $opts->{'runtime'}->{'last_parent_check'} > 5) {
-          if (getppid() eq "1") {
-            kill_gently ($pid, {
-              'first_kill_type' => 'process_group',
-              'final_kill_type' => 'process_group',
-              'wait_time' => $opts->{'terminate_wait_time'}
-              });
-            $parent_died = 1;
-          }
-
-          $opts->{'runtime'}->{'last_parent_check'} = $now;
-        }
-      }
-
-      # user specified timeout
-      if ($opts->{'timeout'}) {
-        if ($now - $start_time > $opts->{'timeout'}) {
-          kill_gently ($pid, {
-            'first_kill_type' => 'process_group',
-            'final_kill_type' => 'process_group',
-            'wait_time' => $opts->{'terminate_wait_time'}
-            });
-          $child_timedout = 1;
-        }
-      }
-
-      # give OS 10 seconds for correct return of waitpid,
-      # kill process after that and finish wait loop;
-      # shouldn't ever happen -- remove this code?
-      if ($got_sig_child) {
-        if ($now - $got_sig_child > 10) {
-          print STDERR "waitpid did not return -1 for 10 seconds after SIG_CHLD, killing [$pid]\n";
-          kill (-9, $pid);
-          $child_finished = 1;
-        }
-      }
-
-      if ($got_sig_quit) {
-#        Yandex::Tools::do_log("ending process group $pid", {'stderr' => 1});
-        kill_gently ($pid, {
-          'first_kill_type' => 'process_group',
-          'final_kill_type' => 'process_group',
-          'wait_time' => $opts->{'terminate_wait_time'}
-          });
-        $child_finished = 1;
-      }
-
-      my $waitpid = waitpid($pid, WNOHANG);
-
-      # child finished, catch it's exit status
-      if ($waitpid ne 0 && $waitpid ne -1) {
-        $child_exit_code = $? >> 8;
-      }
-
-      if ($waitpid eq -1) {
-        $child_finished = 1;
-        next;
-      }
-
-      # child -> parent simple internal communication protocol
-      while (my $l = <$child_info_socket>) {
-        if ($l =~ /^spawned ([0-9]+?)\n(.*?)/so) {
-          $child_child_pid = $1;
-          $l = $2;
-        }
-        if ($l =~ /^reaped ([0-9]+?)\n(.*?)/so) {
-          $child_child_pid = undef;
-          $l = $2;
-        }
-        if ($l =~ /^[\d]+ killed with ([0-9]+?)\n(.*?)/so) {
-          $child_killed_by_signal = $1;
-          $l = $2;
-        }
-      }
-
-      while (my $l = <$child_stdout_socket>) {
-        if (!$opts->{'discard_output'}) {
-          $child_stdout .= $l;
-          $child_merged .= $l;
-        }
-
-        if ($opts->{'stdout_handler'} && ref($opts->{'stdout_handler'}) eq 'CODE') {
-          $opts->{'stdout_handler'}->($l);
-        }
-      }
-      while (my $l = <$child_stderr_socket>) {
-        if (!$opts->{'discard_output'}) {
-          $child_stderr .= $l;
-          $child_merged .= $l;
-        }
-
-        if ($opts->{'stderr_handler'} && ref($opts->{'stderr_handler'}) eq 'CODE') {
-          $opts->{'stderr_handler'}->($l);
-        }
-      }
-
-      Time::HiRes::usleep(1);
-    }
-
-    # $child_pid_pid is not defined in two cases:
-    #  * when our child was killed before
-    #    it had chance to tell us the pid
-    #    of the child it spawned. we can do
-    #    nothing in this case :(
-    #  * our child successfully reaped its child,
-    #    we have nothing left to do in this case
-    #
-    # defined $child_pid_pid means child's child
-    # has not died but nobody is waiting for it,
-    # killing it brutally.
-    #
-    if ($child_child_pid) {
-      kill_gently($child_child_pid);
-    }
-
-    # in case there are forks in child which
-    # do not forward or process signals (TERM) correctly
-    # kill whole child process group, effectively trying
-    # not to return with some children or their parts still running
-    #
-    # to be more accurate -- we need to be sure
-    # that this is process group created by our child
-    # (and not some other process group with the same pgid,
-    # created just after death of our child) -- fortunately
-    # this might happen only when process group ids
-    # are reused quickly (there are lots of processes
-    # spawning new process groups for example)
-    #
-    if ($opts->{'clean_up_children'}) {
-      kill(-9, $pid);
-    }
-
-#    print "child $pid finished\n";
-
-    close($child_stdout_socket);
-    close($child_stderr_socket);
-    close($child_info_socket);
-
-    my $o = {
-      'stdout' => $child_stdout,
-      'stderr' => $child_stderr,
-      'merged' => $child_merged,
-      'timeout' => $child_timedout ? $opts->{'timeout'} : 0,
-      'exit_code' => $child_exit_code,
-      'parent_died' => $parent_died,
-      'killed_by_signal' => $child_killed_by_signal,
-      'child_pgid' => $pid,
-      };
-
-    my $err_msg = "";
-    if ($o->{'exit_code'}) {
-      $err_msg .= "exited with code [$o->{'exit_code'}]\n";
-    }
-    if ($o->{'timeout'}) {
-      $err_msg .= "ran more than [$o->{'timeout'}] seconds\n";
-    }
-    if ($o->{'parent_died'}) {
-      $err_msg .= "parent died\n";
-    }
-    if ($o->{'stdout'} && !$opts->{'non_empty_stdout_ok'}) {
-      $err_msg .= "stdout:\n" . $o->{'stdout'} . "\n";
-    }
-    if ($o->{'stderr'}) {
-      $err_msg .= "stderr:\n" . $o->{'stderr'} . "\n";
-    }
-    if ($o->{'killed_by_signal'}) {
-      $err_msg .= "killed by signal [" . $o->{'killed_by_signal'} . "]\n";
-    }
-    $o->{'err_msg'} = $err_msg;
-
-    if ($orig_sig_child) {
-      $SIG{'CHLD'} = $orig_sig_child;
-    }
-    else {
-      delete($SIG{'CHLD'});
-    }
-
-    return $o;
-  }
-  else {
-    Yandex::Tools::die("cannot fork: $!") unless defined($pid);
-
-    # create new process session for open3 call,
-    # so we hopefully can kill all the subprocesses
-    # which might be spawned in it (except for those
-    # which do setsid theirselves -- can't do anything
-    # with those)
-
-    POSIX::setsid() || Yandex::Tools::die("Error running setsid: " . $!);
-
-    if ($opts->{'child_BEGIN'} && ref($opts->{'child_BEGIN'}) eq 'CODE') {
-      $opts->{'child_BEGIN'}->();
-    }
-
-    close($child_stdout_socket);
-    close($child_stderr_socket);
-    close($child_info_socket);
-
-    my $child_exit_code;
-
-    # allow both external programs
-    # and internal perl calls
-    if (!ref($cmd)) {
-      $child_exit_code = open3_run($cmd, {
-        'parent_info' => $parent_info_socket,
-        'parent_stdout' => $parent_stdout_socket,
-        'parent_stderr' => $parent_stderr_socket,
-        'child_stdin' => $opts->{'child_stdin'},
-        });
-    }
-    elsif (ref($cmd) eq 'CODE') {
-      $child_exit_code = $cmd->({
-        'opts' => $opts,
-        'parent_info' => $parent_info_socket,
-        'parent_stdout' => $parent_stdout_socket,
-        'parent_stderr' => $parent_stderr_socket,
-        'child_stdin' => $opts->{'child_stdin'},
-        });
-    }
-    else {
-      print $parent_stderr_socket "Invalid command reference: " . ref($cmd) . "\n";
-      $child_exit_code = 1;
-    }
-
-    close($parent_stdout_socket);
-    close($parent_stderr_socket);
-    close($parent_info_socket);
-
-    if ($opts->{'child_END'} && ref($opts->{'child_END'}) eq 'CODE') {
-      $opts->{'child_END'}->();
-    }
-
-    exit $child_exit_code;
-  }
-}
+sub install_layered_signal { return IPC::Cmd::install_layered_signal(@_); }
+sub kill_gently { return IPC::Cmd::kill_gently(@_); }
+sub open3_run { return IPC::Cmd::open3_run(@_); }
+sub run_forked { return IPC::Cmd::run_forked(@_); }
 
 sub debug2 {
   my ($text, $opts) = @_;
